/**
 * ldminfo - Part of the Linux-NTFS project.
 *
 * Copyright (C) 2001 Richard Russon <ldm@flatcap.org>
 *
 * Documentation is available at http://linux-ntfs.sourceforge.net/ldm
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program (in the main directory of the Linux-NTFS source
 * in the file COPYING); if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "ldminfo.h"

int * hardsect_size[255];
int * blk_size     [255];
int * blksize_size [255];

struct hd_struct hd      [256];
int              hd_sizes[256];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,2)
struct gendisk   hd_gendisk = { 3, "hd", 6, 64, hd, hd_sizes };
#else
struct gendisk   hd_gendisk = { 3, "hd", 6, hd, hd_sizes };
#endif


int ldm_parse_vmdb(const u8 *buffer, struct vmdb *vm);

/**
 * dump_info - Display a list of partitions, a la fdisk
 */
void dump_info (char *name, int maxminor)
{
	int i;
	int numeric;

	name = basename (name);
	numeric = isdigit (name[strlen (name) - 1]);

	printf ("Device       | Offset Bytes    Sectors     Mb | Size   Bytes    Sectors     Mb\n"
		"-------------+--------------------------------+-------------------------------\n");
	for (i = 0; i < 256; i++) {
		char buf[16];
		int minor = i % maxminor;
		long long start;
		long long size;

		if (hd[i].nr_sects == 0)
			continue;

		if (minor)
			if (numeric)
				sprintf (buf, "%.10sp%d", name, minor);
			else
				sprintf (buf, "%.10s%d", name, minor);
		else
			sprintf (buf, "%.10s", name);

		start = hd[i].start_sect; start <<= 9;
		size  = hd[i].nr_sects;   size  <<= 9;
		printf ("%-12.12s | %12lld %10lld %6lld | %12lld %10lld %6lld\n", buf, start, start>>9, start>>20, size, size>>9, size>>20);
	}
	printf ("\n");
}

/**
 * ldm_partition2
 *
 * This is here so we can extract the structs from the kernel part
 */
int ldm_partition2 (struct gendisk *hd, struct block_device *bdev, struct ldmdb *ldb,
		    int first_minor)
{
	struct vblk   *disk;
	unsigned long base;
	int err = -1;

	if (!ldm_validate_partition_table (bdev))
		return -1;

	if (!ldm_validate_privheads (bdev, 0, &ldb->ph, hd, first_minor))
		return -1;

	/* All further references are relative to base (database start). */
	base = ldb->ph.config_start;

	/* Parse and check tocs and vmdb. */
	if (!ldm_validate_tocblocks (bdev, base, ldb) ||
	    !ldm_validate_vmdb      (bdev, base, ldb))
	    	return -1;

	if (!ldm_create_partition (hd, first_minor++, base, ldb->ph.config_size))
		goto free_buffers;

	if (!ldm_get_vblks (bdev, base, ldb)) {
		printk (LDM_CRIT "get vblks failed\n");
		goto free_buffers;
	}

	disk = ldm_get_disk_objid (ldb);
	if (!disk) {
		printk ("can't find this disk\n");
		goto free_buffers;
	}

	// FIXME first_sector / 0?
	if (ldm_create_data_partitions (hd, 0, first_minor, ldb, disk))
		err = 1;

free_buffers:
	//FIXME no buffers to free
	return err;
}

/**
 * main - ldminfo entry point
 */
int main (int argc, char *argv[])
{
	int minor = 64;
	kdev_t drive = mk_kdev(3,minor);	/* hdb */
	int a;
	int maxminor = 64;
	int info  = 0;
	int dump  = 0;
	int copy  = 0;
	int help  = 0;
	int ver   = 0;
	struct block_device bdev;

	for (a = 1; a < argc; a++) {
		if	(strcmp (argv[a], "--info")    == 0) info++;
		else if	(strcmp (argv[a], "--dump")    == 0) dump++;
		else if (strcmp (argv[a], "--copy")    == 0) copy++;
		else if (strcmp (argv[a], "--debug")   == 0) debug++;
		else if (strcmp (argv[a], "--help")    == 0) help++;
		else if (strcmp (argv[a], "--version") == 0) ver++;
		else continue;
		argv[a][0] = 0;
	}

	/*
	if (help || ver) {
		printf ("\nGDB is free software, covered by the GNU General Public License, and you are\n");
		printf ("welcome to change it and/or distribute copies of it under certain conditions.\n");
	}
	*/

	if (help || (argc - info - dump - copy - debug) < 2) {
		printf ("\nUsage:\n    %s [options] device ...\n", basename (argv[0]));
		printf ("\nOptions:\n"
			"    --info     A concise list of partitions (default)\n"
			"    --dump     The contents of the database in detail\n"
			"    --copy     Write the database to a file\n"
			"    --debug    Display lots of debugging information\n"
			"    --version  display the version number\n"
			"    --help     Show this short help\n\n");
		return 1;
	}

	if (ver) {
		printf ("\nldminfo v0.0.6\n\n");
		return 1;
	}

	for (a = 1; a < argc; a++) {
		long long size;
		struct ldmdb     ldb; //FIXME move to heap

		if (!argv[a][0])
			continue;

		device = open64 (argv[a], O_RDONLY);
		if (device < 0) {
			printf ("Couldn't open device: %s\n", argv[a]);
			break;
		}

		size = lseek64 (device, 0, SEEK_END);
		if (size < 0) {
			printf ("Seek failed for device: %s\n", argv[a]);
			break;
		}

		if (copy) {
			copy_database (argv[a], device, size);
			goto close;
		}

		/* Initialize vblk list in ldmdb struct */
		INIT_LIST_HEAD(&ldb.v_dgrp);
		INIT_LIST_HEAD(&ldb.v_disk);
		INIT_LIST_HEAD(&ldb.v_volu);
		INIT_LIST_HEAD(&ldb.v_comp);
		INIT_LIST_HEAD(&ldb.v_part);

		memset (hardsect_size, 0, sizeof (hardsect_size));
		memset (blk_size,      0, sizeof (blk_size));
		memset (blksize_size,  0, sizeof (blksize_size));
		memset (hd,            0, sizeof (hd));
		memset (hd_sizes,      0, sizeof (hd_sizes));

		hd_gendisk.part[minor].nr_sects = size >> 9;
		hd_gendisk.sizes[minor] = size >> (BLOCK_SIZE_BITS - 9);
		blk_size[major(drive)] = hd_gendisk.sizes;
		bdev.bd_dev = kdev_val(drive);	 /* hack (bd_dev isn't a kdev_t) */

		if (ldm_partition2 (&hd_gendisk, &bdev, &ldb, minor + 1) != 1) {
			printf ("Something went wrong, skipping device '%s'\n", argv[a]);
			continue;
		}

#if 1
		if (dump)
			dump_database (argv[a], &ldb);
		else
			dump_info     (argv[a], maxminor);
#endif

		ldm_free_vblks(&ldb.v_dgrp);
		ldm_free_vblks(&ldb.v_disk);
		ldm_free_vblks(&ldb.v_volu);
		ldm_free_vblks(&ldb.v_comp);
		ldm_free_vblks(&ldb.v_part);
close:
		close (device);
		device = 0;
	}

	if (device)
		close (device);

	//printf ("%d/%d\n", ldm_a, ldm_f);
	return 0;
}


